//include some standard libraries
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <iostream>

//include OpenGL libraries
#include <GL/glew.h>
#include <GLFW/glfw3.h>

//include some custom code files
#include "glfunctions.h" //include all OpenGL stuff
#include "Shader.h" // class to compile shaders
#include "imageloader.h"

#define TINYOBJLOADER_IMPLEMENTATION
#include "tiny_obj_loader.h"
//#include "imageloader.h"
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>


using namespace std;
using namespace glm;

//global variables to help us do things
int g_ViewportWidth = 512;
int g_ViewportHeight = 512; // Default window size, in pixels
double mouse_x, mouse_y; //variables storing mouse position
double pos_x, pos_y, pos_z;
const vec3 g_backgroundColor(0.2f, 0.2f, 0.2f); // background colour - a GLM 3-component vector


GLuint g_simpleShader = 0; //shader identifier
GLuint g_MilkShader = 1;
GLuint g_Vao = 0; //vao
GLuint g_NumTriangles = 0; //  Numbre of triangles we are painting.
GLuint texture_milkyway;
GLuint texture_earth;

vec3 camera_position(0.0, 0.0, -17.0);
vec3 camera_target(0.0, 0.0, -20.0);
vec2 last_target(0.0, 0.0);


float camaraX = 0;
float camaraZ = 3;
//int g_ViewportWidth = 1024; 
//int g_ViewportHeight = 512;

glm::vec3 eye = glm::vec3(camaraX, 0.0f, camaraZ);
glm::vec3 center = glm::vec3(0.0f, 0.0f, -1.0f);
glm::vec3 up = glm::vec3(0.0f, 1.0f, 0.0f);



struct Thing
{
	std::string asset_path = "";
	GLuint simpleShader = 0; //shader identifier
	GLuint Vao = 0; //vao
	GLuint NumTriangles = 0; //  Numbre of triangles we are painting.
	vec4  color;

};

std::vector<Thing> things;

// ------------------------------------------------------------------------------------------
// This function manually creates a square geometry (defined in the array vertices[])
// ------------------------------------------------------------------------------------------
void load()
{
	/*string basepath = "assets/";
	string inputfile = basepath + "plane.obj";
	vector< tinyobj::shape_t > shapes2;
	vector< tinyobj::material_t > materials2;
	string err2;
	bool ret = tinyobj::LoadObj(shapes2, materials2, err2, inputfile.c_str(), basepath.c_str());*/

	for (Thing& t : things)
	{
		std::string basepath = "assets/";
		std::string inputfile = basepath + "sphere.obj";
		std::vector< tinyobj::shape_t > shapes;
		std::vector< tinyobj::material_t > materials;
		std::string err;
		bool ret = tinyobj::LoadObj(shapes, materials, err, inputfile.c_str(), basepath.c_str());

		if (!err.empty())
		{ // `err` may contain warning message.    
			std::cerr << err << std::endl;
		}


		//**********************
		// CODE TO SET GEOMETRY
		//**********************

		//the positions of the corner of the square, each a 3-component vector, x, y, z
		/*const GLfloat vertices[] = {
			1.0f, -1.0f, 0.0f,
			1.0f,  1.0f, 0.0f,
			-1.0f,  1.0f, 0.0f,
			-1.0f, -1.0f, 0.0f
			};

		//we assign a colour to each corner (each colour is RGB)
		const GLfloat colors[] = {
			//r    g     b
			1.0f, 0.0f, 0.0f,
			1.0f, 0.0f, 0.0f,
			1.0f, 0.0f, 0.0f,
			1.0f, 0.0f, 0.0f
			};

		// The index buffer references the vertices we paint, in order
		//here we have two triangles
		const GLuint indices[] = { 0, 1, 2, 0, 2, 3 };*/



		//**********************
		// CODE TO LOAD EVERYTHING INTO MEMORY
		//**********************

		//load the shader
		Shader simpleShader("src/shader.vert", "src/shader.frag");
		t.simpleShader = simpleShader.program;

		// Create the VAO where we store all geometry (stored in g_Vao)
		t.Vao = gl_createAndBindVAO();

		//create vertex buffer for positions, colors, and indices, and bind them to shader
		gl_createAndBindAttribute(shapes[0].mesh.positions.data(), shapes[0].mesh.positions.size() * sizeof(float), t.simpleShader, "a_vertex", 3);
		gl_createAndBindAttribute(shapes[0].mesh.texcoords.data(), shapes[0].mesh.texcoords.size() * sizeof(float), t.simpleShader, "a_uv", 2);
		//coje el array de indices, tienes tanta memoria entonces copiala.
		gl_createIndexBuffer(shapes[0].mesh.indices.data(), shapes[0].mesh.indices.size() * sizeof(unsigned int)); //el unsigned int ens posa tota la part negativa a la part positiva i per tant tenim el doble de bits en la part positiva ja que nomes treballarem amb ells


		/*//create vertex buffer for positions, colors, and indices, and bind them to shader
		gl_createAndBindAttribute(shapes2[0].mesh.positions.data(), shapes2[0].mesh.positions.size() * sizeof(float), g_simpleShader, "a_vertex", 3);
		gl_createAndBindAttribute(shapes2[0].mesh.texcoords.data(), shapes2[0].mesh.texcoords.size() * sizeof(float), g_simpleShader, "a_uv", 2);
		//coje el array de indices, tienes tanta memoria entonces copiala.
		gl_createIndexBuffer(shapes2[0].mesh.indices.data(), shapes2[0].mesh.indices.size() * sizeof(unsigned int)); //el unsigned int ens posa tota la part negativa a la part positiva i per tant tenim el doble de bits en la part positiva ja que nomes treballarem amb ells*/


		//unbind everything
		gl_unbindVAO();

		Image* imagemilkyway = loadBMP("milkyway.bmp");
		glGenTextures(1, &texture_milkyway);
		glBindTexture(GL_TEXTURE_2D, texture_milkyway);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, //target
			0, //level, = 0 base, no mipmap
			GL_RGB, //the way opengl stores the data
			imagemilkyway->width, //width
			imagemilkyway->height, //height
			0,
			GL_RGB, //the format of the original data
			GL_UNSIGNED_BYTE, //type of data
			imagemilkyway->pixels); //pointer to start of data


		Image* imageearth = loadBMP("earthmap1k.bmp");
		glGenTextures(1, &texture_earth);
		glBindTexture(GL_TEXTURE_2D, texture_earth);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, //target
			0, //level, = 0 base, no mipmap
			GL_RGB, //the way opengl stores the data
			imageearth->width, //width
			imageearth->height, //height
			0,
			GL_RGB, //the format of the original data
			GL_UNSIGNED_BYTE, //type of data
			imageearth->pixels); //pointer to start of data

		//store number of triangles (use in draw())
		//t.NumTriangles = sizeof(indices) / (sizeof(GLuint) * 3);
		t.NumTriangles = shapes[0].mesh.indices.size() / 3; //num de cares 
		//g_NumTriangles = shapes2[0].mesh.indices.size() / 3; //num de cares 
	}
}

// ------------------------------------------------------------------------------------------
// This function actually draws to screen and called non-stop, in a loop
// ------------------------------------------------------------------------------------------
void draw_milkyway() 
{
	//milk
	GLuint model_loc2 = glGetUniformLocation(g_simpleShader, "u_model");  //find u_model in shader   
	GLuint u_texture2 = glGetUniformLocation(g_simpleShader, "u_texture");
	GLuint projection_loc = glGetUniformLocation(g_simpleShader, "u_projection");
	GLuint u_view = glGetUniformLocation(g_simpleShader, "u_view");

	mat4 projection_matrix = glm::perspective(60.0f, (float)(1024 / 512), 0.1f, 50.0f);
	mat4 model_matrix = translate(mat4(1.0f), vec3(-0.5, -0.5, -6.0));   //translate identity in -0.5x -0.6y    
	vec3 eye_vector(1.0f, -3.0f, 1.0f);
	vec3 up_vector(0.0f, 1.0f, 0.0f);
	vec3 center_vector(0.0f, 0.0f, 0.0f);

	mat4 view_matrix = glm::lookAt(
		eye,
		center,
		up
	);

	glUniform1i(u_texture2, 1);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, texture_milkyway);

	//bind the geometry
	gl_bindVAO(g_Vao);

	// Draw to screen
	glDrawElements(GL_TRIANGLES, 3 * g_NumTriangles, GL_UNSIGNED_INT, 0);

	//unbind vao and shader
	gl_unbindVAO();
	glUseProgram(0);

}

void draw()
{
	//clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glClearColor(.3, .3, .3, 1.0);


	for (Thing& t : things)
	{
		// activate shader
		glUseProgram(t.simpleShader);

		//amb aixo podem canviar la posicio de la tetera en la camara.
		
		//earth
		GLuint model_loc = glGetUniformLocation(t.simpleShader, "u_model");  //find u_model in shader   
		GLuint u_texture = glGetUniformLocation(g_simpleShader, "u_texture");
		GLuint projection_loc = glGetUniformLocation(t.simpleShader, "u_projection");
		GLuint u_view = glGetUniformLocation(t.simpleShader, "u_view");


		mat4 model_matrix = translate(mat4(1.0f), vec3(-0.5, -0.5, -6.0));   //translate identity in -0.5x -0.6y    
		//create projection matrix and pass to shader 
		mat4 projection_matrix = perspective(
			60.0f, // Field of view   
			1.0f,  // Aspect ratio   
			0.1f,  // near plane (distance from camera)   
			50.0f  // Far plane (distance from camera) 
		);
		vec3 eye_vector(1.0f, -3.0f, 1.0f);
		vec3 up_vector(0.0f, 1.0f, 0.0f);
		vec3 center_vector(0.0f, 0.0f, 0.0f);

		mat4 view_matrix = glm::lookAt(
			eye,
			center,
			up
		);

		//upload the data to the shader
		glUniformMatrix4fv(model_loc, 1, GL_FALSE, glm::value_ptr(model_matrix));  
		glUniformMatrix4fv(projection_loc, 1, GL_FALSE, glm::value_ptr(projection_matrix));
		glUniformMatrix4fv(u_view, 1, GL_FALSE, glm::value_ptr(view_matrix));

		//bind the sampler to texture unit
		glUniform1i(u_texture, 0);

		//activate texture unit 0 and bind our texture object
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texture_earth);
		

		//bind the geometry
		gl_bindVAO(t.Vao);

		// Draw to screen
		glDrawElements(GL_TRIANGLES, 3 * t.NumTriangles, GL_UNSIGNED_INT, 0);
	
		//unbind vao and shader
		gl_unbindVAO();
		glUseProgram(0);


	}


}

// ------------------------------------------------------------------------------------------
// This function is called every time you press a screen
// ------------------------------------------------------------------------------------------
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	//quit
	if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) glfwSetWindowShouldClose(window, 1);
	if (key == GLFW_KEY_R && action & (GLFW_PRESS | GLFW_REPEAT)) load();
	if (key == GLFW_KEY_W && (action == GLFW_PRESS || action == GLFW_REPEAT)) camaraZ -= 0.01f;
	if (key == GLFW_KEY_A && (action == GLFW_PRESS || action == GLFW_REPEAT)) camaraX -= 0.01f;
	if (key == GLFW_KEY_S && (action == GLFW_PRESS || action == GLFW_REPEAT)) camaraZ += 0.01f;
	if (key == GLFW_KEY_D && (action == GLFW_PRESS || action == GLFW_REPEAT)) camaraX += 0.01f;
}


float pitchx;
float yawy;
int directionx = cos(glm::radians(pitchx)) * cos(glm::radians(yawy));
int directiony = sin(glm::radians(pitchx));
int directionz = cos(glm::radians(pitchx)) * sin(glm::radians(yawy));
float lastX = 400, lastY = 300;
bool firstMouse = true;

void mouse_callback(GLFWwindow* window, double xpos, double ypos)
{
	if (firstMouse)
	{
		lastX = xpos;
		lastY = ypos;
		firstMouse = false;
	}

	float xoffset = xpos - lastX;
	float yoffset = lastY - ypos;
	lastX = xpos;
	lastY = ypos;

	float sensitivity = 0.05;
	xoffset *= sensitivity;
	yoffset *= sensitivity;

	yawy += xoffset;
	pitchx += yoffset;

	if (pitchx > 89.0f)
		pitchx = 89.0f;
	if (pitchx < -89.0f)
		pitchx = -89.0f;

	glm::vec3 front;
	front.x = cos(glm::radians(yawy)) * cos(glm::radians(pitchx));
	front.y = sin(glm::radians(pitchx));
	front.z = sin(glm::radians(yawy)) * cos(glm::radians(pitchx));
	center = glm::normalize(front);
}


// ------------------------------------------------------------------------------------------
// This function is called every time you click the mouse
// ------------------------------------------------------------------------------------------
void mouse_button_callback(GLFWwindow* window, int button, int action, int mods) {
	if (button == GLFW_MOUSE_BUTTON_LEFT) {
		cout << "Left mouse down at" << mouse_x << ", " << mouse_y << endl;
	}
}

int main(void)
{
	//setup window and boring stuff, defined in glfunctions.cpp
	GLFWwindow* window;
	if (!glfwInit())return -1;
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	window = glfwCreateWindow(g_ViewportWidth, g_ViewportHeight, "Hello OpenGL!", NULL, NULL);
	if (!window) { glfwTerminate();	return -1; }
	glfwMakeContextCurrent(window);
	glewExperimental = GL_TRUE;
	glewInit();

	//input callbacks
	glfwSetKeyCallback(window, key_callback);
	glfwSetMouseButtonCallback(window, mouse_button_callback);
	glfwSetInputMode(window, GLFW_STICKY_KEYS, 1);
	glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
	glfwSetCursorPosCallback(window, mouse_callback);


	Thing  tetera;
	tetera.asset_path = "teapot_small.obj";
	tetera.color = vec4(1, 0, 0, 1);
	things.push_back(tetera);



	//load all the resources
	load();

	// Loop until the user closes the window
	while (!glfwWindowShouldClose(window))
	{
		draw_milkyway();
		draw();

		// Swap front and back buffers
		glfwSwapBuffers(window);

		// Poll for and process events
		glfwPollEvents();


		float currentFrame = glfwGetTime();

		//mouse position must be tracked constantly (callbacks do not give accurate delta)
		glfwGetCursorPos(window, &mouse_x, &mouse_y);

	}

	//terminate glfw and exit
	glfwTerminate();
	return 0;
}


